/*******************************************************************************
 *
 * File Name: neuralNetworkFormation.cpp
 * Author: Richard Baxter
 * Project: Advanced Neural Network (ANN) - Fractal Based Neural Network
 * Project Version: 1b20b 12-Oct-07
 * Comments: NB CHECK THIS new requirement; each subnet must have at least 3 layers
 *
 *******************************************************************************/

 //IMPORTANT CODING NOTE - on 10-dec-06/1b6b I have started to remove the neuronReference class as circular referencing does not appear to be allowed in C++
//NB when create NeuronList class change all referecnes to "...List->frontNeuronReferences.." to "...List->neuronReferences" as Neurons when declared for the purpose of creating neuron lists only have their frontNeuronReferences used

#include <ctime>
#include <cstdlib>	//for randon number generation
#include <cmath>

#include <string>
#include <iostream>
#include <fstream>
using namespace std;

#include "neuralNetworkFormation.h"


//const int numberOfInputNeurons = NUM_INPUT_NEURONS_TH;
//const int numberOfOutputNeurons = NUM_OUTPUT_NEURONS_TH;
//const int maxFolds = MAX_NUM_FOLDS_TH;
//const int maxSetsOfData = MAX_SETS_OF_DATA_TH;

static Neuron * formNondistinctLayeredNeuralNetwork(Neuron * firstInputNeuronInNetwork, long numberOfInputNeurons, long numberOfOutputNeurons, long numberOfLayers, double probabilityOfSubnetCreation, long maxNumberOfRecursiveSubnets, bool useSubnetDependentNumberOfLayers, double subnetNumberOfLayersModifier, int subnetDivergenceType, double meanSubnetDivergenceFactor, double probabilityOfNeuronConnectionExclusivity);
	static Neuron * fillNondistinctHiddenLayer(Neuron * firstNeuronInCurrentLayer, long numberOfInputNeurons, long numberOfOutputNeurons, long numberOfNeuronsInCurrentLayer, long currentNumberOfLayers, long numberOfLayers, double probabilityOfSubnetCreation, long maxNumberOfRecursiveSubnets, long currentNumberOfRecursiveSubnets, bool useSubnetDependentNumberOfLayers, double subnetNumberOfLayersModifier, int subnetDivergenceType, double meanSubnetDivergenceFactor, double probabilityOfNeuronConnectionExclusivity);
		static long calculateNumberOfNeuronsInNewFrontLayer(long numberOfInputNeurons, long numberOfOutputNeurons, long numberOfNeuronsInCurrentLayer, long currentNumberOfLayers, long numberOfLayers, double meanSubnetDivergenceFactor, int subnetDivergenceType);
			static double advancedDivergenceFactor(long currentNumberOfLayersFromNearestEndPoint, long numberOfLayers);
		static long calcNumberOfLayersInSubnet(long numberOfInputNeurons, long numberOfOutputNeurons, long currentNumberOfRecursiveSubnets, long maxNumberOfRecursiveSubnets, bool useSubnetDependentNumberOfLayers, double subnetNumberOfLayersModifier);


static long IDCounter;
static long orderIDCounter;
static long layerIDCounter;
static long subnetIDCounter;

/************************************************************ Form Neural Network Routines ******************************************************/

/*forms a 3 layered net similar to that used in original c project*/
//assumes firstInputNeuron and firstOutputNeuron have already been filled
void formSimpleNeuralNet(Neuron * firstInputNeuron, Neuron * firstOutputNeuron, long numberOfInputNeurons, long numberOfHiddenNeurons, long numberOfOutputNeurons)
{
	IDCounter = DEFAULT_FIRST_OUTPUT_NEURON_ID+1;
	subnetIDCounter = 1;

	/*NB simple neural network heursitc;
	number of hidden neurons = average of input and output neurons
	*/

	Neuron * firstHiddenNeuron = new Neuron(IDCounter, orderIDCounter, layerIDCounter, subnetIDCounter);
	IDCounter++;

	/*Create all Neurons in Simply Layered Net*/

	//initiate Input Layer in Neural Network
	orderIDCounter = 1;
	layerIDCounter = 10;
	Neuron * currentNeuronReference = firstInputNeuron;

	for(long i = 0; i < numberOfInputNeurons; i++)
	{
		currentNeuronReference->firstNeuronInFrontLayer = firstHiddenNeuron;
		currentNeuronReference->hasFrontLayer = true;

		Neuron * newNeuron = new Neuron(IDCounter, orderIDCounter, layerIDCounter, subnetIDCounter);
		IDCounter++;
		orderIDCounter++;
		currentNeuronReference->nextNeuron = newNeuron;
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}
	IDCounter--;

	//initiate Hidden Layer in Neural Network
	orderIDCounter = 1;
	layerIDCounter++;
	currentNeuronReference = firstHiddenNeuron;
	for(long i = 0; i < numberOfHiddenNeurons; i++)
	{
		currentNeuronReference->firstNeuronInFrontLayer = firstOutputNeuron;
		currentNeuronReference->firstNeuronInBackLayer = firstInputNeuron;
		currentNeuronReference->hasFrontLayer = true;
		currentNeuronReference->hasBackLayer = true;

		Neuron * newNeuron = new Neuron(IDCounter, orderIDCounter, layerIDCounter, subnetIDCounter);
		IDCounter++;
		orderIDCounter++;
		currentNeuronReference->nextNeuron = newNeuron;
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}
	IDCounter--;

	//initiate Output Layer in Neural Network
	orderIDCounter = 1;
	layerIDCounter++;
	currentNeuronReference = firstOutputNeuron;
	for(long i = 0; i < numberOfOutputNeurons; i++)
	{
		currentNeuronReference->firstNeuronInBackLayer = firstHiddenNeuron;
		currentNeuronReference->hasBackLayer = true;

		Neuron * newNeuron = new Neuron(IDCounter, orderIDCounter, layerIDCounter, subnetIDCounter);
		IDCounter++;
		orderIDCounter++;
		currentNeuronReference->nextNeuron = newNeuron;
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}
	IDCounter--;

	/*Link all Neurons in Simply Layered Net - fill frontNeuronReferences and BackNeuronReferences for all neurons*/

	//Link ALL input and hidden layer neurons neurons (fill in frontNeuronReferences/BackNeuronReferences list for all relevant neurons)

	Neuron * currentBackLayerNeuron;
	Neuron * currentForwardLayerNeuron;
	currentBackLayerNeuron = firstInputNeuron;
	for(long i = 0; i < numberOfInputNeurons; i++)
	{
		currentForwardLayerNeuron = firstHiddenNeuron;
		for(long j = 0; j < numberOfHiddenNeurons; j++)
		{
			NeuronConnection * newNeuronConnection = new NeuronConnection();
			currentForwardLayerNeuron->numBackNeuronConnections = numberOfInputNeurons;
			currentBackLayerNeuron->numFrontNeuronConnections = numberOfHiddenNeurons;
			currentForwardLayerNeuron->backNeuronReferences[i] = currentBackLayerNeuron;
			currentBackLayerNeuron->frontNeuronReferences[j] = currentForwardLayerNeuron;
			currentForwardLayerNeuron->backNeuronRefConnections[i] = newNeuronConnection;
			currentBackLayerNeuron->frontNeuronRefConnections[j] = newNeuronConnection;
			currentForwardLayerNeuron = currentForwardLayerNeuron->nextNeuron;
		}
		currentBackLayerNeuron = currentBackLayerNeuron->nextNeuron;
	}


	//Link ALL hidden and output layer neurons neurons (fill in frontNeuronReferences/BackNeuronReferences list for all relevant neurons)
	currentBackLayerNeuron = firstHiddenNeuron;
	for(long i = 0; i < numberOfHiddenNeurons; i++)
	{
		currentForwardLayerNeuron = firstOutputNeuron;
		for(long j = 0; j < numberOfOutputNeurons; j++)
		{
			NeuronConnection * newNeuronConnection = new NeuronConnection();
			currentForwardLayerNeuron->numBackNeuronConnections = numberOfHiddenNeurons;
			currentBackLayerNeuron->numFrontNeuronConnections = numberOfOutputNeurons;
			currentForwardLayerNeuron->backNeuronReferences[i] = currentBackLayerNeuron;
			currentBackLayerNeuron->frontNeuronReferences[j] = currentForwardLayerNeuron;
			currentForwardLayerNeuron->backNeuronRefConnections[i] = newNeuronConnection;
			currentBackLayerNeuron->frontNeuronRefConnections[j] = newNeuronConnection;

			currentForwardLayerNeuron = currentForwardLayerNeuron->nextNeuron;
		}
		currentBackLayerNeuron = currentBackLayerNeuron->nextNeuron;
	}

}



















/************************************************************ End Form Neural Network Routines ******************************************************/



















	//this is currently only psuedo code

/************************************************************ Form Advanced Neural Network Routines ******************************************************/

Neuron * formNeuralNet(Neuron * firstInputNeuron, long numberOfInputNeurons, long numberOfOutputNeurons, double subnetNumberOfLayersModifier, int subnetDivergenceType, double meanSubnetDivergenceFactor, double probabilityOfNeuronConnectionExclusivity)
{
	#ifdef DEBUG
	cout << "\nstatic Neuron * formNeuralNet()\n" << endl;
	#endif

	long numberOfLayers = calcNumberOfLayersInSubnet(numberOfInputNeurons, numberOfOutputNeurons, 1, 1, false, subnetNumberOfLayersModifier);
	cout << "Number of Layers = " << numberOfLayers << "\n" << endl;


	IDCounter = DEFAULT_FIRST_INPUT_NEURON_ID+1;
	orderIDCounter = 1;
	layerIDCounter = 1;
	subnetIDCounter = 1;

	Neuron * firstOutputNeuronInNetwork;

	double probabilityOfSubnetCreation = 0.0;	/*no divergence = no ANN*/
	long maxNumberOfRecursiveSubnets = 0;		/*no recursive subnets = no ANN*/

	firstOutputNeuronInNetwork = formNondistinctLayeredNeuralNetwork(firstInputNeuron, numberOfInputNeurons, numberOfOutputNeurons, numberOfLayers, probabilityOfSubnetCreation, maxNumberOfRecursiveSubnets, false, subnetNumberOfLayersModifier, subnetDivergenceType, meanSubnetDivergenceFactor, probabilityOfNeuronConnectionExclusivity);

	return firstOutputNeuronInNetwork;
}


#ifdef ANN
Neuron * formAdvancedNeuralNetwork(Neuron * firstInputNeuron, long numberOfInputNeurons, long numberOfOutputNeurons, bool useSubnetDependentNumberOfLayers, double probabilityOfSubnetCreation, long maxNumberOfRecursiveSubnets, double subnetNumberOfLayersModifier, int subnetDivergenceType, double meanSubnetDivergenceFactor, double probabilityOfNeuronConnectionExclusivity)
{
	#ifdef DEBUG
	cout << "\nstatic Neuron * formAdvancedNeuralNetwork()\n" << endl;
	#endif

	long numberOfLayers = calcNumberOfLayersInSubnet(numberOfInputNeurons, numberOfOutputNeurons, 1, maxNumberOfRecursiveSubnets, useSubnetDependentNumberOfLayers, subnetNumberOfLayersModifier);
	cout << "Number of Layers = " << numberOfLayers << "\n" << endl;

	IDCounter = DEFAULT_FIRST_INPUT_NEURON_ID+1;
	orderIDCounter = 1;
	layerIDCounter = 1;
	subnetIDCounter = 1;

	Neuron * firstOutputNeuronInNetwork;

	firstOutputNeuronInNetwork = formNondistinctLayeredNeuralNetwork(firstInputNeuron, numberOfInputNeurons, numberOfOutputNeurons, numberOfLayers, probabilityOfSubnetCreation, maxNumberOfRecursiveSubnets, useSubnetDependentNumberOfLayers, subnetNumberOfLayersModifier, subnetDivergenceType, meanSubnetDivergenceFactor, probabilityOfNeuronConnectionExclusivity);

	return firstOutputNeuronInNetwork;
}
#endif

static Neuron * formNondistinctLayeredNeuralNetwork(Neuron * firstInputNeuronInNetwork, long numberOfInputNeurons, long numberOfOutputNeurons, long numberOfLayers, double probabilityOfSubnetCreation, long maxNumberOfRecursiveSubnets, bool useSubnetDependentNumberOfLayers, double subnetNumberOfLayersModifier, int subnetDivergenceType, double meanSubnetDivergenceFactor, double probabilityOfNeuronConnectionExclusivity)
{
	#ifdef DEBUG
	cout << "\nstatic Neuron * formNondistinctLayeredNeuralNetwork()\n" << endl;
	#endif

	Neuron * firstOutputNeuronInNetwork;

	//seed random number generator
	srand(static_cast<unsigned>(time(0)));


	//initiate Input Layer in Neural Network
	Neuron * currentNeuronReference = firstInputNeuronInNetwork;

	for(long i = 0; i < numberOfInputNeurons; i++)
	{
		Neuron * newNeuron = new Neuron(IDCounter, orderIDCounter, layerIDCounter, subnetIDCounter);
		IDCounter++;
		orderIDCounter++;

		currentNeuronReference->nextNeuron = newNeuron;
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}
	IDCounter--;

	firstOutputNeuronInNetwork = fillNondistinctHiddenLayer(firstInputNeuronInNetwork, numberOfInputNeurons, numberOfOutputNeurons, numberOfInputNeurons, 1, numberOfLayers, probabilityOfSubnetCreation, maxNumberOfRecursiveSubnets, 1, useSubnetDependentNumberOfLayers, subnetNumberOfLayersModifier, subnetDivergenceType, meanSubnetDivergenceFactor, probabilityOfNeuronConnectionExclusivity);

	return firstOutputNeuronInNetwork;
}


static Neuron * fillNondistinctHiddenLayer(Neuron * firstNeuronInCurrentLayer, long numberOfInputNeurons, long numberOfOutputNeurons, long numberOfNeuronsInCurrentLayer, long currentNumberOfLayers, long numberOfLayers, double probabilityOfSubnetCreation, long maxNumberOfRecursiveSubnets, long currentNumberOfRecursiveSubnets, bool useSubnetDependentNumberOfLayers, double subnetNumberOfLayersModifier, int subnetDivergenceType, double meanSubnetDivergenceFactor, double probabilityOfNeuronConnectionExclusivity)
{
	subnetIDCounter = currentNumberOfRecursiveSubnets;

	#ifdef DEBUG
	cout << "\nstatic Neuron * fillNondistinctHiddenLayer()\n" << endl;
	#endif


	Neuron * firstOutputNeuronInNetwork;

	//Create neurons for the new front layer

	long numberOfNeuronsInNewFrontLayer;

	#define IRRELEVANT 0.0

	numberOfNeuronsInNewFrontLayer = calculateNumberOfNeuronsInNewFrontLayer(numberOfInputNeurons, numberOfOutputNeurons, numberOfNeuronsInCurrentLayer, currentNumberOfLayers, numberOfLayers, meanSubnetDivergenceFactor, subnetDivergenceType);

	#ifdef DEBUG
	cout << "numberOfNeuronsInNewFrontLayer = " << numberOfNeuronsInNewFrontLayer << endl;
	#endif

	layerIDCounter = currentNumberOfLayers+1;
	orderIDCounter=1;
	Neuron * firstNeuronInNewFrontLayer = new Neuron(IDCounter, orderIDCounter, layerIDCounter, subnetIDCounter);
	IDCounter++;
	orderIDCounter++;
	Neuron * currentNeuronReference = firstNeuronInNewFrontLayer;
	//fill in network description variables for new front layer (CHECK THIS - this is incomplete)
	for(long i=0; i<numberOfNeuronsInNewFrontLayer; i++)
	{
		currentNeuronReference->firstNeuronInBackLayer = firstNeuronInCurrentLayer;
		currentNeuronReference->hasBackLayer = true;
		if(currentNumberOfLayers != (numberOfLayers-1))
		{
			currentNeuronReference->hasFrontLayer = true;	//CHECK THIS - need to update neuralNetworkUpdateAlgorithm.cpp accordingly
		}


		Neuron * newNeuron = new Neuron(IDCounter, orderIDCounter, layerIDCounter, subnetIDCounter);
		IDCounter++;
		orderIDCounter++;
		currentNeuronReference->nextNeuron = newNeuron;
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}
	IDCounter--;

	currentNeuronReference = firstNeuronInCurrentLayer;
	long numberOfNeuronsOnCurrentLayer = 0;
	//fill in network description variables for current layer (CHECK THIS - this is incomplete)
	while(currentNeuronReference -> nextNeuron != NULL)
	{
		currentNeuronReference->hasFrontLayer = true;
		currentNeuronReference->firstNeuronInFrontLayer = firstNeuronInNewFrontLayer;

		currentNeuronReference = currentNeuronReference->nextNeuron;
		numberOfNeuronsOnCurrentLayer++;
	}


	//Modify associations/links based upon layers of intermediate/hidden neurons
	//if probabilityOfNeuronConnectionExclusivity is 1.0, link ALL current/back and front layer neurons neurons (fill in frontNeuronReferences/BackNeuronReferences list for all relevant neurons)


	if(compareDoubles(probabilityOfNeuronConnectionExclusivity, 1.0))
	{
		Neuron * currentBackLayerNeuron = firstNeuronInCurrentLayer;
		for(long i = 0; i < numberOfNeuronsOnCurrentLayer; i++)
		{
			Neuron * currentForwardLayerNeuron = firstNeuronInCurrentLayer->firstNeuronInFrontLayer;
			for(long j = 0; j < numberOfNeuronsInNewFrontLayer; j++)
			{
				NeuronConnection * newNeuronConnection = new NeuronConnection();

				currentForwardLayerNeuron->numBackNeuronConnections = numberOfNeuronsOnCurrentLayer;
				currentBackLayerNeuron->numFrontNeuronConnections = numberOfNeuronsInNewFrontLayer;
				currentForwardLayerNeuron->backNeuronReferences[i] = currentBackLayerNeuron;
				currentBackLayerNeuron->frontNeuronReferences[j] = currentForwardLayerNeuron;
				currentForwardLayerNeuron->backNeuronRefConnections[i] = newNeuronConnection;
				currentBackLayerNeuron->frontNeuronRefConnections[j] = newNeuronConnection;

				currentForwardLayerNeuron = currentForwardLayerNeuron->nextNeuron;
			}
			currentBackLayerNeuron = currentBackLayerNeuron->nextNeuron;
		}
	}
	else
	{
		//now record neuron connections for front layer neurons
		Neuron * currentBackLayerNeuron = firstNeuronInCurrentLayer;
		while(currentBackLayerNeuron->nextNeuron != NULL)			//for(long i = 0; i < numberOfNeuronsOnCurrentLayer; i++)
		{
			long numConnectionsWithFrontLayerNeurons = 0;
			Neuron * currentForwardLayerNeuron = firstNeuronInCurrentLayer->firstNeuronInFrontLayer;
			while(currentForwardLayerNeuron->nextNeuron != NULL)		//for(long j = 0; j < numberOfNeuronsInNewFrontLayer; j++)
			{

				srand( (unsigned)time(NULL) );	//seeds randomness
				double randomNumberBetween0And1 = ((double(abs((short)rand())))/(ABS_OF_SHORT));
				//cout << "randomNumberBetween0And1 = " << randomNumberBetween0And1 << endl;

				if(randomNumberBetween0And1 < probabilityOfNeuronConnectionExclusivity)			//if(randomNumberBetween0And1 > 0)	//always diverge
				{
					NeuronConnection * newNeuronConnection = new NeuronConnection();

					currentBackLayerNeuron->frontNeuronReferences[numConnectionsWithFrontLayerNeurons] = currentForwardLayerNeuron;
					currentBackLayerNeuron->frontNeuronRefConnections[numConnectionsWithFrontLayerNeurons] = newNeuronConnection;

					numConnectionsWithFrontLayerNeurons++;
				}

				currentForwardLayerNeuron = currentForwardLayerNeuron->nextNeuron;
			}
			currentBackLayerNeuron->numFrontNeuronConnections = numConnectionsWithFrontLayerNeurons;

			currentBackLayerNeuron = currentBackLayerNeuron->nextNeuron;
		}

		//now record neuron connections for front layer neurons [using only the information previously recorded for the back layer connections]
		Neuron * currentForwardLayerNeuron = firstNeuronInCurrentLayer->firstNeuronInFrontLayer;
		while(currentForwardLayerNeuron->nextNeuron != NULL)		//for(long j = 0; j < numberOfNeuronsInNewFrontLayer; j++)
		{
			long numConnectionsWithBackLayerNeurons = 0;
			Neuron * currentBackLayerNeuron = firstNeuronInCurrentLayer;
			while(currentBackLayerNeuron->nextNeuron != NULL)			//for(long i = 0; i < numberOfNeuronsOnCurrentLayer; i++)
			{
				for(long q = 0; q < currentBackLayerNeuron->numFrontNeuronConnections; q++)
				{
					if(currentBackLayerNeuron->frontNeuronReferences[q]->id == currentForwardLayerNeuron->id)
					{
						currentForwardLayerNeuron->backNeuronReferences[numConnectionsWithBackLayerNeurons] = currentBackLayerNeuron;
						currentForwardLayerNeuron->backNeuronRefConnections[numConnectionsWithBackLayerNeurons] = currentBackLayerNeuron->frontNeuronRefConnections[q];
						numConnectionsWithBackLayerNeurons++;
					}
				}
				currentBackLayerNeuron = currentBackLayerNeuron->nextNeuron;

			}
			currentForwardLayerNeuron->numBackNeuronConnections = numConnectionsWithBackLayerNeurons;

			currentForwardLayerNeuron = currentForwardLayerNeuron->nextNeuron;
		}
	}

#ifdef ANN
	Neuron * currentBackLayerNeuron = firstNeuronInCurrentLayer;
	for(long i = 0; i < numberOfNeuronsOnCurrentLayer; i++)
	{
		//create a subnet if necessary
		Neuron * currentNeuron = currentBackLayerNeuron;		//define the neuron which will become a subnet

		if(currentNeuron->hasBackLayer)	//can only create a subnet if the neuron candidate has a back layer
		{
			srand( (unsigned)time(NULL) );	//seeds randomness

			//cout << "\ncurrentNumberOfRecursiveSubnets = " << currentNumberOfRecursiveSubnets << endl;
			//cout << "maxNumberOfRecursiveSubnets = " << maxNumberOfRecursiveSubnets << endl;

			if(currentNumberOfRecursiveSubnets < maxNumberOfRecursiveSubnets)
			{
				double randomNumberBetween0And1 = ((double(abs((short)rand())))/(ABS_OF_SHORT));
				//cout << "randomNumberBetween0And1 = " << randomNumberBetween0And1 << endl;

				if(randomNumberBetween0And1 < probabilityOfSubnetCreation)
				{
					#ifdef PROPER_LARGE_NUMBER_OF_NN_INPUTS
					if(abs(currentNeuron->numBackNeuronConnections - currentNeuron->numFrontNeuronConnections) >= 2)
					{//only create a subnet if the difference in the number of neurons between the front and the back layers is at least 2
					#endif

						orderIDCounter=1;
						layerIDCounter=1;
						subnetIDCounter = currentNumberOfRecursiveSubnets+1;
						Neuron * firstNeuronInBackLayerOfSubnet = new Neuron(IDCounter, orderIDCounter, layerIDCounter, subnetIDCounter);
						IDCounter++;
						orderIDCounter++;
						Neuron * currentNeuronInBackLayerOfSubnet = firstNeuronInBackLayerOfSubnet;
						currentNeuron->numNeuronsInBackLayerOfSubnet = currentNeuron->numBackNeuronConnections;
						currentNeuron->numNeuronsInFrontLayerOfSubnet =  currentNeuron->numFrontNeuronConnections;
						long q;

						for(q = 0; q < currentNeuron->numBackNeuronConnections; q++)
						{
									//OLD: now replace input layer of subnet with currentNeuron->backNeuronReferences [this is not necessary as 1-1 mapping will be assumed here during the neural network training / update algorithm]
											//copyNeuronBPropertiesIntoNeuronA(currentNeuronInBackLayerOfSubnet, currentNeuron->backNeuronReferences[q]);
										//assume [1-1] map back neurons of currentNeuron with input neurons of subnet		[CHECK THIS - nb this 1-1 mapping must be maintained in the neuronNetworkUpdateAlgorithm, not just in the neuralNetworkFormation]

							currentNeuronInBackLayerOfSubnet->hasBackLayer = false;	//CHECK THIS - need to update neuralNetworkUpdateAlgorithm.cpp accordingly
							currentNeuronInBackLayerOfSubnet->hasFrontLayer = true;

							Neuron * newNeuron = new Neuron(IDCounter, orderIDCounter, layerIDCounter, subnetIDCounter);
							IDCounter++;
							orderIDCounter++;
							currentNeuronInBackLayerOfSubnet->nextNeuron = newNeuron;
							currentNeuronInBackLayerOfSubnet = currentNeuronInBackLayerOfSubnet->nextNeuron;
						}
						IDCounter--;

						currentNeuron->isSubnet = true;
						currentNeuron->firstNeuronInBackLayerOfSubnet = firstNeuronInBackLayerOfSubnet;

						Neuron * firstNeuronInFrontLayerOfSubnet;
						long numberOfHiddenLayersInSubnet = calcNumberOfLayersInSubnet(currentNeuron->numNeuronsInBackLayerOfSubnet, currentNeuron->numNeuronsInFrontLayerOfSubnet, currentNumberOfRecursiveSubnets, maxNumberOfRecursiveSubnets, useSubnetDependentNumberOfLayers, subnetNumberOfLayersModifier);
						#define NONLINEARDIVERGENCEFACTOR (1)

						#ifdef DEBUG
						cout << "\nAbout to perform ANN recursion in fillNondistinctHiddenLayer ... " << endl;
						cout << "currentNeuron Neuron ID = " << currentNeuron->id << endl;
						cout << "currentNeuron Neuron ID Order = " << currentNeuron->orderID << endl;
						cout << "currentNeuron Neuron ID Layer = " << currentNeuron->layerID << endl;
						cout << "currentNeuron Neuron ID Subnet = " << currentNeuron->subnetID << endl;
						cout << "firstInputNeuronInNetwork = firstNeuronInBackLayerOfSubnet = ... " << endl;
						cout << "numberOfInputNeurons = " << currentNeuron->numNeuronsInBackLayerOfSubnet << endl;
						cout << "numberOfOutputNeurons = " <<  currentNeuron->numNeuronsInFrontLayerOfSubnet << endl;
						cout << "numberOfNeuronsInCurrentLayer = " << currentNeuron->numNeuronsInBackLayerOfSubnet << endl;
						cout << "currentNumberOfLayers = " << 1 << endl;
						cout << "numberOfHiddenLayersInSubnet = " << numberOfHiddenLayersInSubnet << endl;
						cout << "probabilityOfSubnetCreation*NONLINEARDIVERGENCEFACTOR = " << probabilityOfSubnetCreation*NONLINEARDIVERGENCEFACTOR << endl;
						cout << "maxNumberOfRecursiveSubnets = " << maxNumberOfRecursiveSubnets << endl;
						cout << "currentNumberOfRecursiveSubnets+1 = " << currentNumberOfRecursiveSubnets+1 << endl;
						#endif


						firstNeuronInFrontLayerOfSubnet = fillNondistinctHiddenLayer(firstNeuronInBackLayerOfSubnet, currentNeuron->numNeuronsInBackLayerOfSubnet, currentNeuron->numNeuronsInFrontLayerOfSubnet, currentNeuron->numNeuronsInBackLayerOfSubnet, 1, numberOfHiddenLayersInSubnet, probabilityOfSubnetCreation*NONLINEARDIVERGENCEFACTOR, maxNumberOfRecursiveSubnets, currentNumberOfRecursiveSubnets+1, useSubnetDependentNumberOfLayers, subnetNumberOfLayersModifier, subnetDivergenceType, meanSubnetDivergenceFactor, probabilityOfNeuronConnectionExclusivity);


						currentNeuron->firstNeuronInFrontLayerOfSubnet = firstNeuronInFrontLayerOfSubnet;
							//now replace output layer of subnet with frontLayerOfSubnet [this is not necessary as 1-1 mapping will be assumed here during the neural network training / update algorithm]
							//[1-1] map forward layer neurons of currentNeuron with output neurons of subnet	[CHECK THIS - nb this 1-1 mapping must be maintained in the neuronNetworkUpdateAlgorithm, not just in the neuralNetworkFormation]
											//reset layer/subnet IDs after ANN recursion;	//this is not currently required as layerIDCounter/subnetIDCounter are not used again in the current execution of this function
						layerIDCounter = currentNumberOfLayers+1;			//reset layer to new front layer
						subnetIDCounter = currentNumberOfRecursiveSubnets;

					#ifdef PROPER_LARGE_NUMBER_OF_NN_INPUTS
					}
					#endif
				}
			}
		}
		currentBackLayerNeuron = currentBackLayerNeuron->nextNeuron;
	}
#endif


	if(currentNumberOfLayers == (numberOfLayers-1))
	{//CHECK THIS - nb this needs to imply that numberOfOutputNeurons = number of neurons in current layer
		if(numberOfOutputNeurons != numberOfNeuronsInNewFrontLayer)
		{
			cout << "Error: error with NN heuristic; (currentNumberOfLayers == (numberOfLayers-1)) && (numberOfNeuronsInNewFrontLayer != numberOfOutputNeurons)" << endl;
			exit(0);
		}

		//reached final output layer
		firstOutputNeuronInNetwork = firstNeuronInNewFrontLayer;
	}
	else
	{
		//now recurse
		firstOutputNeuronInNetwork = fillNondistinctHiddenLayer(firstNeuronInNewFrontLayer, numberOfInputNeurons, numberOfOutputNeurons, numberOfNeuronsInNewFrontLayer, (currentNumberOfLayers+1), numberOfLayers, probabilityOfSubnetCreation, maxNumberOfRecursiveSubnets, currentNumberOfRecursiveSubnets, useSubnetDependentNumberOfLayers, subnetNumberOfLayersModifier, subnetDivergenceType, meanSubnetDivergenceFactor, probabilityOfNeuronConnectionExclusivity);
	}

	return firstOutputNeuronInNetwork;
}

static long calcNumberOfLayersInSubnet(long numberOfInputNeurons, long numberOfOutputNeurons, long currentNumberOfRecursiveSubnets, long maxNumberOfRecursiveSubnets, bool useSubnetDependentNumberOfLayers, double subnetNumberOfLayersModifier)
{
	long numberOfLayersInSubnet;

	if(useSubnetDependentNumberOfLayers == false)
	{
		if(numberOfInputNeurons > numberOfOutputNeurons)
		{
			numberOfLayersInSubnet = (long)((double)(numberOfInputNeurons - numberOfOutputNeurons) * subnetNumberOfLayersModifier) + 1;
		}
		else
		{
			numberOfLayersInSubnet = (long)((double)(numberOfOutputNeurons - numberOfInputNeurons) * subnetNumberOfLayersModifier) + 1;
		}
	}
	else
	{
		if(numberOfInputNeurons > numberOfOutputNeurons)
		{
			numberOfLayersInSubnet = (long)((double)(numberOfInputNeurons - numberOfOutputNeurons) * subnetNumberOfLayersModifier * ((double)currentNumberOfRecursiveSubnets/(double)maxNumberOfRecursiveSubnets)) + 1;
		}
		else
		{
			numberOfLayersInSubnet = (long)((double)(numberOfOutputNeurons - numberOfInputNeurons) * subnetNumberOfLayersModifier * ((double)currentNumberOfRecursiveSubnets/(double)maxNumberOfRecursiveSubnets)) + 1;
		}
	}

	return numberOfLayersInSubnet;
}


/************************************************************ End Form Advanced Neural Network Routines ******************************************************/















































static long calculateNumberOfNeuronsInNewFrontLayer(long numberOfInputNeurons, long numberOfOutputNeurons, long numberOfNeuronsInCurrentLayer, long currentNumberOfLayers, long numberOfLayers, double meanSubnetDivergenceFactor, int subnetDivergenceType)
{

	long numberOfNeuronsInNewFrontLayer;

#ifdef DEBUG
	cout << "\nstatic long calculateNumberOfNeuronsInNewFrontLayer()\n" << endl;
	cout << "currentNumberOfLayers = " << currentNumberOfLayers << endl;
	cout << "numberOfLayers = " << numberOfLayers << endl;
	cout << "numberOfInputNeurons = " << numberOfInputNeurons << endl;
	cout << "numberOfOutputNeurons = " << numberOfOutputNeurons << endl;
	cout << "((currentNumberOfLayers+1)/(numberOfLayers+1)) = " << (double(currentNumberOfLayers)/double(numberOfLayers-1)) << endl;
#endif


	long currentNumberBasicNeuronVariation;
	long currentNumberDivergenceNeurons;

	if(numberOfInputNeurons > numberOfOutputNeurons)
	{
		currentNumberBasicNeuronVariation = -long(double(numberOfInputNeurons - numberOfOutputNeurons) * (double(currentNumberOfLayers)/double(numberOfLayers-1)));
	}
	else
	{
		currentNumberBasicNeuronVariation = long(double(numberOfOutputNeurons - numberOfInputNeurons) * (double(currentNumberOfLayers)/double(numberOfLayers-1)));
	}



	if(subnetDivergenceType == BASIC_SUBNET_DIVERGENCE)
	{
		currentNumberDivergenceNeurons = 0;
	}
	else if((subnetDivergenceType == INTERMEDIATE_SUBNET_DIVERGENCE) || (subnetDivergenceType == ADVANCED_SUBNET_DIVERGENCE))
	{

		if(subnetDivergenceType == INTERMEDIATE_SUBNET_DIVERGENCE)
		{
			if(currentNumberOfLayers < (numberOfLayers/2))
			{
				currentNumberDivergenceNeurons = numberOfInputNeurons * meanSubnetDivergenceFactor * currentNumberOfLayers;
			}
			else
			{
				currentNumberDivergenceNeurons = (numberOfInputNeurons * meanSubnetDivergenceFactor * (numberOfLayers/2)) - (numberOfInputNeurons * meanSubnetDivergenceFactor * (currentNumberOfLayers-(numberOfLayers/2)));
			}
		}
		else if(subnetDivergenceType == ADVANCED_SUBNET_DIVERGENCE)
		{

			if(currentNumberOfLayers < (numberOfLayers/2))
			{
				currentNumberDivergenceNeurons = numberOfInputNeurons * meanSubnetDivergenceFactor * advancedDivergenceFactor(currentNumberOfLayers, numberOfLayers) * currentNumberOfLayers;
			}
			else
			{
				currentNumberDivergenceNeurons = numberOfInputNeurons * meanSubnetDivergenceFactor * advancedDivergenceFactor((numberOfLayers - currentNumberOfLayers), numberOfLayers) * currentNumberOfLayers;
			}
		}

	}
	else
	{
		cout << "illegal subnetDivergenceType detected" << endl;
		exit(0);
	}


	numberOfNeuronsInNewFrontLayer = numberOfInputNeurons + currentNumberDivergenceNeurons + currentNumberBasicNeuronVariation;



		//BASIC_SUBNET_DIVERGENCE NB static Divergence Rate: assuming convergence not divergence, each hidden layer has one less neuron, numberOfNeuronsInNewFrontLayer = numberOfInputNeurons - long(double(numberOfInputNeurons - numberOfOutputNeurons) * (double(currentNumberOfLayers)/double(numberOfLayers-1)));
		//INTERMEDIATE_SUBNET_DIVERGENCE NB meanSubnetDivergenceFactor is less than 1 always; Eg 0.1 or 10% more neurons than input layer per layer




	return numberOfNeuronsInNewFrontLayer;

	//numberOfNeuronsInNewFrontLayer is calculated based upon the description of divergence for any given subnet

	/*Basic Subnet divergence	/	USE_BASIC_SUBNET_DIVERGENCE

		/   On On On  \
	   /   Hn Hn Hn Hn \
	  /  Hn Hn Hn Hn Hn  \
	 /  In In In In In In  \

	using simple layer hidden Neuron heuristic; nb  numberOfLayers = (numberOfInputNeurons + numberOfOutputNeurons)/2

	*/


	/*Intermediate Subnet divergence	/	USE_INTERMEDIATE_SUBNET_DIVERGENCE

					/\
				  /    \
				/        \
			  /	           \
			/                \
		  /                    \
		/                        \

	Input 1                        Output 1
	Input 2                        Output 2
	Input 3                        Output 3
	Input 4                        Output 4
	Input N                        Output N

	   \                         /
		 \                     /
		   \                 /
			 \             /
			   \         /
				 \     /
				   \ /



	NB divergence = meanSubnetDivergenceFactor until 1/2 numberOfLayers is reached,  afterwards meanSubnetDivergenceFactor = -meanSubnetDivergenceFactor  (until number of outputNeurons is reached).
	this implies a ~ diamond shapped (linearily increasing/decreasing) resulting neural net*/


	/* Advanced Subnet divergence	/	USE_ADVANCED_SUBNET_DIVERGENCE

						/\
					   /  \
					  /    \
					/	     \
				  /            \
			   /                  \
			/                        \

		Input 1                        Output 1
		Input 2                        Output 2
		Input 3                        Output 3
		Input 4                        Output 4
		Input N                        Output N

		 \	                         /
			\                     /
			   \               /
				  \          /
					\      /
					 \    /
					  \  /
					   \/

	NB divergence = meanSubnetDivergenceFactor*f(currentNumberOfLayers)  until 1/2 numberOfLayers is reached,  afterwards meanSubnetDivergenceFactor = -meanSubnetDivergenceFactor*f(currentNumberOfLayers)  (until number of outputNeurons is reached)
	depending upon function f this could implies a ~ diamond shapped (non-linearily increasing/decreasing) resulting neural net
	another possibility is that f(currentNumberOfLayers) could be rand()

   */

}



static double advancedDivergenceFactor(long currentNumberOfLayersFromNearestEndPoint, long numberOfLayers)
{
	double advancedDivergenceFactor;

#ifdef USE_ADVANCED_RANDOM_SUBNET_DIVERGENCE
	//seed random number generator
	srand(static_cast<unsigned>(time(0)));
	advancedDivergenceFactor = rand();
#else
	advancedDivergenceFactor = currentNumberOfLayersFromNearestEndPoint/numberOfLayers;
#endif

	return advancedDivergenceFactor;
}



#define DOUBLE_MIN_PRECISION 0.00001
bool compareDoubles(double a, double b)
{
	bool result;
	if((a < (b+DOUBLE_MIN_PRECISION)) && (a > (b-DOUBLE_MIN_PRECISION)))
	{
		result = true;
	}
	else
	{
		result = false;
	}

	return result;
}





